// PGA.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"


/*************************************************************************************
GENETSKI ALGORITAM ZA RESAVANJE PROBLEMA STRONG MINIMUM ENERGY TOPOLOGY (SMETP)
ZA SLUCAJ SA SIMETRICNIM PODACIMA

Implementiran je GA za CPU, kao i za GPU (CUDA) arhitekturu. 
Autor: Aleksandar Kartelj, Matematicki Fakultet, Beograd
**************************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>

/*velicina populacije, uzimamo umnozak broja 32, zbog velicine warp-a*/
#define POPSIZE 128
/*broj elitnih jedinki*/
#define NELITE 16
/*broj igraca u turnirskoj selekciji*/
#define NCOMPET 8
/*pocetna verovatnoca mutacije, moze se povecavati kasnije*/
#define MUTATIONPROB 0.005
/*verovatnoca ukrstanja*/
#define CROSSOVERPROB 0.85
/*beskonacno*/
#define INF 10000000
/*minimizacija = 1 / maksimizacija =0*/
#define MINIMIZATION 1

/*informacije o problemu, puni se na startu*/
problem *prob;
/*populacija*/
chrom popcurrent[POPSIZE];
/*u selection-u nam treba kopija zbog razmestanja*/
chrom poptemp[POPSIZE];
/*pomocni niz, za proveru stroge povezanosti 
u CUDA implementaciji bi trebalo da ide za svaki device, 
ili jos bolje, prebaciti na bitovske operatore, tako da 
bi onda trebalo samo nekoliko promenljivih, u zavisnosti
od broja cvorova*/
char *visited;
/*lista indeksa suseda: na i-tom mestu pocinju susedi
i-tog cvora (neighbourIdx)*/
int *nIdx;
/*tezina funkcije cilja, napravljene su vestacke funkcije
tako da budu slozenosti objvalhardness*O(n^3), specijalno
ako je vrednost 0, radi se originalna funkcija cilja*/
int objvalhardness;
/*gustina grafa, koristi se pri inicijalizaciji
vrednosti. Sto je manje, graf ce biti redji*/
float graphdensity;

/*glavna funkcija*/
void ga(char *path, int randSeed, int num, int objhard, float grphdens)                                    
{
	/*broj iteracija*/
	int i,j;
	time_t t;
	clock_t start, found,end;
	double lastBest;
	int bestRepeated;
	double modifiedMutationProb=MUTATIONPROB;
	char outPath[255]="./out.dat";
	char rSeedStr[10];
	char numOfGenStr[10];
	objvalhardness=objhard;
	graphdensity=grphdens;
	FILE *out;
	/*pocinje izvrsavanje*/
	time(&t);
	start=clock();
	loadInstance(path);
	//sprintf(outPath,"./tmp_%d_%d.dat",randSeed,num);
	out=fopen(outPath,"a");
	if(out==NULL){
		fprintf(stderr, "Greska pri otvaranju fajla %s\n", outPath);
		exit(0);
	}
	printf("\nSTRONG MINIMUM ENERGY TOPOLOGY PROBLEM (SMETP)\n");
	printf("\nBroj iteracija:  %d\n",num);
	srand(randSeed);
	/*inicijalizacija pocetne populacije*/
	init();
	evpop();
	order(0);
	lastBest=INF;
	for(i=0;i<num;i++)                            
	{
		if((MINIMIZATION && popcurrent[0].fit<lastBest) || (!MINIMIZATION && popcurrent[0].fit>lastBest))
		{
			lastBest=popcurrent[0].fit;
			found=clock();
			bestRepeated=1;
			modifiedMutationProb=MUTATIONPROB;
		}else{
			bestRepeated++;
			if(bestRepeated%1000==0)
				modifiedMutationProb+=0.005;
		}
		printf("Najbolje:\t%f\ti = %d\n",lastBest,i);  
		/*selekcija*/
		selection();
		/*ukrstanje*/
		crossover(1);   
		/*mutacija nad linkovima*/
		mutation();
		/*racunamo f-je cilja i fitness-e za sve jedinke*/
		evpop();         
		/*uredjivanje*/
		order(0);
		/*ako su se pojavili duplikati, izbacujemo ih jer narusavaju
		diverzitet genetskog koda*/
		replaceEqual();
		/*sortiramo ponovo*/
		order(0);
	}
	end=clock();
	/*zavrsilo se izvrsavanje, ispis...*/
	fprintf(out,"-----------------------------------------------------------\
				\ntime:\t\t %s\
				\npath:\t\t%s\
				\nseed:\t\t%d\
				\nmaxGen:\t%d\
				\npopSize:\t%d\
				\nobjHard:\t%d\
				\nnelite:\t%d\
				\nstartMut:\t%.4f\
				\nendMut:\t%.4f\
				\ncrossProb:\t%.4f\
				\ntournSize:\t%d\
				\ngraphDens:\t%.4f\
				\nbest:\t\t%d\
				\nsec:\t\t%.4f\
				\ntotSec:\t%.4f\n",ctime(&t),strrstr(path,"/"),randSeed,num, POPSIZE,objvalhardness, NELITE,MUTATIONPROB,
				modifiedMutationProb,CROSSOVERPROB,NCOMPET,grphdens,(int)lastBest,
				((double) (found - start)) / CLOCKS_PER_SEC,((double) (end - start)) / CLOCKS_PER_SEC);
	fclose(out);
	freeAll();
}                                                      


void loadInstance(char *path){
	FILE *fp;
	int j,a,b,w;
	int neighbNum=0;
	fp=fopen(path,"r");
	if(fp == NULL){
		printf("Ne moze da se otvori fajl %s \n",path);
		exit(0);
	}
	prob=(problem*)malloc(sizeof(problem));
	if(prob==NULL){
		fprintf(stderr, "Greska pri alokaciji problem strukture!");
		exit(0);
	}
	fscanf(fp, "%d %d",&(prob->nn), &(prob->nl));
	/*grane su u fajlu zapisane orijentisano, nas interesuje samo
	simetricni slucaj, tako da ih je duplo manje*/
	prob->nl=(prob->nl)/2;
	visited=(char*)malloc(sizeof(char)*prob->nn);
	/*pakujemo indekse susedstva*/
	nIdx=(int*)malloc(sizeof(int)*(prob->nn-1));
	neighbNum=prob->nn-1;
	nIdx[0]=0;
	for(j=1; j<prob->nn-1; j++){
		nIdx[j]=(nIdx[j-1]+neighbNum);
		neighbNum--;
	}

	prob->l = (link*)malloc(prob->nl * sizeof(link));
	if(prob->l == NULL){ 
		fprintf(stderr, "Greska pri alokaciji linkova");
		exit (0); 
	}
	j=0;
	while(j<prob->nl){
		fscanf(fp,"%d %d %d",&a,&b,&w); 
		if(a<b){
			prob->l[j].from=a-1;
			prob->l[j].to=b-1;
			prob->l[j].weight=w;
			j++;
		}
	}
	fclose(fp);
}

void init(){
	int i,j;
	/*alociramo hromozome*/
	for(j=0;j<POPSIZE;j++)                  
	{
		popcurrent[j].z=(int*)malloc(prob->nn*sizeof(int));
		/*alociramo bitove za linkove, dodajemo jedan vise zbog terminirajuce nule*/
		popcurrent[j].lbit=(char*)malloc((prob->nl+1)*sizeof(char));

		/*pomocni, treba nam posle za selekciju*/
		poptemp[j].z=(int*)malloc(prob->nn*sizeof(int));
		poptemp[j].lbit=(char*)malloc((prob->nl+1)*sizeof(char));

		if(popcurrent[j].lbit==NULL || popcurrent[j].z==NULL || 
			poptemp[j].lbit==NULL || poptemp[j].z==NULL){
				fprintf(stderr, "Greska pri alokaciji z i lbit nizova");
				exit(0);
		}
		generateRandom(j);
	}	
	/*azuriramo i pomocni niz, samo kloniranje vrednost*/
	fillTempPop();
}

void ev(int j){
	obj(&popcurrent[j]);
	/*fines funkcija jedinke*/
	fitness(&popcurrent[j]);
}

void evpop() 
{
	int j;
	for(j=0;j<POPSIZE;j++)                  
	{
		ev(j);
	}           
}                               

void obj(chrom *chr)   
{
	double sum=0;
	int i;
	if(Pvalid(chr,objvalhardness)){
		for(i=0; i<prob->nn; i++){
			chr->z[i]=0;
		}
		for(i=0; i<prob->nl; i++){
			/*ako je grana pripadajuca, azuriramo energiju senzora
			na toj poziciji (u slucaju da je veca*/
			if(chr->lbit[i]=='1'){
				/*mora da se azurira sa obe strane grane*/
				if(prob->l[i].weight>chr->z[prob->l[i].from])
					chr->z[prob->l[i].from]=prob->l[i].weight;
				if(prob->l[i].weight>chr->z[prob->l[i].to])
					chr->z[prob->l[i].to]=prob->l[i].weight;
			}
		}
		for(i=0; i<prob->nn; i++){
			sum+=(chr->z[i]);
		}
		(*chr).obj=sum;
		(*chr).valid=1;
	}else{
		(*chr).obj=INF;
		(*chr).valid=0;
	}
}                             

int edgeIndex(int i,int j){
	int tidx;
	if(i<j)
		tidx=nIdx[i]+(j-i)-1;
	else if(i>j)
		tidx=nIdx[j]+(i-j)-1;
	else 
		tidx=-1;
	return tidx;
}


void dfs(chrom *chr, int idx){
	int i,eidx;
	visited[idx]=1;
	for(i=0; i<prob->nn; i++){
		/*ako je i==idx, ne proveravamo*/
		if(i==idx)
			continue;
		/*najpre, proveravamo da li postoji grana izmedju dva cvora
		u ovom hromozomu. Znamo tacno gde se nalazi grana za dva proizvoljna cvora
		zato sto su grafovi kompletni, pa smo ih lepo uredili. Uredjeni su ovako: 
		(0 1),(0,2),............,(0,nn-1),
		(1,2),(1,3),...,(1,nn-1),
		..
		(nn-2,nn-1)*/

		eidx=edgeIndex(idx,i);
		if(chr->lbit[eidx]=='1'){
			/*ako postoji grana, onda proveravamo da li smo u 
			tom cvoru vec bili*/
			if(visited[i])
				continue;
			else{
				dfs(chr, i);
			}
		}
	}
}

int bfs(chrom *chr, int idx){
	int *stack;
	int sp=0;
	int i;
	int eidx;
	int hasNext=1;
	int count=1;
	for(int i=0; i<prob->nn; i++)
		visited[i]=0;
	stack=(int*)malloc(sizeof(int)*prob->nn);
	for(i=0; i<prob->nn; i++)
		stack[i]=0;
	if(stack==NULL){
		printf("Nije uspela alokacija steka\n");
		exit(0);
	}
	stack[sp]=idx;
	visited[idx]=1;
	while(sp<count && count<prob->nn){
		for(i=0; i<prob->nn; i++){
			if(!visited[i]){
				eidx=edgeIndex(stack[sp],i);
				if(chr->lbit[eidx]=='1'){
					stack[count++]=i;
					visited[i]=1;
				}
			}
		}
		sp++;
	}
	free(stack);
	return count==prob->nn;
}


/*validno je ukoliko je povezano*/
int valid(chrom *chr){
	return bfs(chr,0);
}

/*samo drugi uslov validnosti, kazemo da 
graf mora da bude povezan, tako da izmedju svaka
dva cvora postoji bar jedan put duzine 3, 
ne zaboraviti prepodesavanje GRAPHDENSITY parametra*/
int Pvalid(chrom *chr, int P){
	if(P>0){
		int complexity=1;
		for(int p=0; p<P;p++)
			complexity*=prob->nn;
		int sum=0;
		for(int i=0; i<complexity; i++){
			if(rand()%2==0)
				sum+=(rand()%100);
			else
				sum-=(rand()%100);
		}
	}
	return valid(chr);
}




int asccmp(const void *v1, const void *v2)
 {
     return (*(chrom *)v1).fit - (*(chrom *)v2).fit;
 }

int desccmp(const void *v1, const void *v2)
 {
     return (*(chrom *)v1).fit - (*(chrom *)v2).fit;
 }

/*uredjivanje*/
void order(int desc) 
{

	int i,j;
	chrom temp;   
	if(desc)
		qsort(popcurrent,POPSIZE,sizeof(popcurrent[0]),desccmp);
	else
		qsort(popcurrent, POPSIZE, sizeof(popcurrent[0]), asccmp);
}   



void fillTempPop(){
	int i,j;
	for(i=0; i<POPSIZE; i++){
		poptemp[i].obj=popcurrent[i].obj;
		poptemp[i].fit=popcurrent[i].fit;
		poptemp[i].valid=popcurrent[i].valid;
		for(j=0; j<prob->nn; j++){
			poptemp[i].z[j]=popcurrent[i].z[j];
		}
		for(j=0; j<prob->nl; j++){
			poptemp[i].lbit[j]=popcurrent[i].lbit[j];
		}
	}
}

/***************************************************************\
selection()

Vrsi selekciju tako sto pronalazi najbolji medju NCOMPET slucajno
izabranih jedinki. Ideja je da se elitne jedinke ostavljaju. 
Nakon toga vrsi se odabir POPSIZE-NELITE jedinki turnirskom 
selekcijom. Na ovaj nacin, odredjene jedinke ce najv. imati
vise od 1 ponavljanja. 
\***************************************************************/
void selection()
{
	int intItem, intChosen, intTemp;
	int intChamp, intCompet;
	int ind[POPSIZE];
	for(intItem=0; intItem<POPSIZE; intItem++){
		ind[intItem]=0;
	}
	/* u svakom koraku, pobednika turnira stavljamo
	na intItem lokaciju*/
	for ( intItem=NELITE; intItem<POPSIZE; intItem++ )
	{
		intChamp= rand()%(POPSIZE-NELITE)+NELITE;
		for ( intChosen=1; intChosen<NCOMPET; intChosen++ )
		{
			intCompet= rand()%(POPSIZE-NELITE)+NELITE;
			if ((MINIMIZATION && popcurrent[intCompet].fit <popcurrent[intChamp].fit)
				|| (!MINIMIZATION && popcurrent[intCompet].fit>popcurrent[intChamp].fit))
				intChamp= intCompet;
		}
		ind[intItem]=intChamp;
	}
	/*pravimo kopiju populacije*/
	fillTempPop();
	/*sada je sve spremno za razmestanje po dobijenim indeksima*/
	for(intItem=NELITE; intItem<POPSIZE; intItem++){
		popcurrent[intItem].obj=poptemp[ind[intItem]].obj;
		popcurrent[intItem].fit=poptemp[ind[intItem]].fit;
		popcurrent[intItem].valid=poptemp[ind[intItem]].valid;
		/*kopriramo i vrednosti cvorovnih i link bitova*/
		for(intTemp=0; intTemp<prob->nn; intTemp++){
			popcurrent[intItem].z[intTemp]=poptemp[ind[intItem]].z[intTemp];
		}
		for(intTemp=0; intTemp<prob->nl; intTemp++){
			popcurrent[intItem].lbit[intTemp]=poptemp[ind[intItem]].lbit[intTemp];
		}
	}

}  /* TournamentSelect */

void replaceEqual(){
	int i,j, equal;
	for(i=1; i<POPSIZE-1; i++){
		/*kandidat za duplikata*/
		if(popcurrent[i].obj==popcurrent[i+1].obj){
			equal=1;
			/*ipak, moraju da se provere i svi bitovi, posto moze da se desi
			da je ista f-ja cilja, a razlicita konfiguracija*/
			for(j=0; j<prob->nl; j++){
				if(popcurrent[i].lbit[j]!=popcurrent[i+1].lbit[j]){
					equal=0;
					break;
				}
			}
			if(equal){
				/*sad podesi novu konfiguraciju za i-ti hromozom*/
				generateRandom(i);
				ev(i);
			}
		}
	}
}

void generateRandom(int j){
	double random;
	int i;
	/*inicijalizujemo linkove (slucajno)*/
	for(i=0;i<prob->nl;i++)         
	{
		/*inicijalizujemo slucajno 0 ili 1 vrednosti*/
		random=rand()*1.0/RAND_MAX; 
		if(random<graphdensity)
			popcurrent[j].lbit[i]='1';
		else
			popcurrent[j].lbit[i]='0';
	}   
	popcurrent[j].lbit[prob->nl]='\0';	
	poptemp[j].lbit[prob->nl]='\0';
}

/* Ne vodi racuna o tome da se neke jedinke ukrstaju vise puta */
void crossover(int type)
{
	int i,j,itm1,itm2,site, i1, i2;
	double crossProb;
	/*pomocni niz za obelezavanje koji je hromozom vec
	ucestvovao u rekombinaciji*/
	int ind[POPSIZE];
	char temp;
	int popsize=POPSIZE;
	/*za sve NEelitne jedinke vrsimo ukrstanje*/
	for(i=0; i<POPSIZE; i++)
		ind[i]=0;
	for(i=NELITE; i<POPSIZE-1; i+=2){
		/*biramo slucajnu tacku ukrstanja*/
		/*razmenjujemo kodove*/
		i1 = rand()%(popsize-NELITE);
		itm1 = NELITE;
		/*biramo onu jedinku koja je prva najbliza
		indeksu i1, a pritom nad njom nije vrsena
		rekombinacija*/
		for(j=0; j<=i1;j++)
		{
			while (ind[itm1]) itm1++;
			itm1++;
		}
		itm1--;
		popsize--;
		ind[itm1] = 1;

		i2 = rand()%(popsize-NELITE);
		itm2 = NELITE;
		for(j=0; j<=i2;j++)
		{
			while (ind[itm2]) itm2++;
			itm2++;
		}
		itm2--;
		popsize--;
		ind[itm2] = 1;

		crossProb=rand()*1.0/RAND_MAX;
		if(crossProb<CROSSOVERPROB){
			if(type==1){
				/*jednopoziciona rekombinacija*/
				site=rand()%prob->nl;
				for(j=site; j<prob->nl; j++){
					temp=popcurrent[itm1].lbit[j];
					popcurrent[itm1].lbit[j]=popcurrent[itm2].lbit[j];
					popcurrent[itm2].lbit[j]=temp;
				}
			}else if(type==INF){
				/*uniformna rekombinacija*/		
				for(j=0; j<prob->nl; j++){
					if((rand()*1.0/RAND_MAX)<0.5){
						temp=popcurrent[itm1].lbit[j];
						popcurrent[itm1].lbit[j]=popcurrent[itm2].lbit[j];
						popcurrent[itm2].lbit[j]=temp;
					}
				}
			}
		}
	}
} 

int frozenBit(int idx){
	for(int i=0; i<POPSIZE-1; i++)
		if(popcurrent[i].lbit[idx]!=popcurrent[i+1].lbit[idx])
			return 0;
	return 1;
}

void toggle(int itm, int idx){
	if (popcurrent[itm].lbit[idx]=='0')
		popcurrent[itm].lbit[idx]='1' ;
	else if (popcurrent[itm].lbit[idx]=='1')
		popcurrent[itm].lbit[idx]='0';
}

void mutation() 
{
	double random;
	int i,j;
	/*na elitne ne primenjujemo mutaciju*/
	for(i=NELITE; i<POPSIZE; i++){
		/*prolazimo kroz celu populaciju*/
		for(j=0; j<prob->nl; j++){
			/*kroz svaki lokus*/
			random=rand()*1.0/RAND_MAX;
			if(random<MUTATIONPROB){
				/*i sa verov. mutiranja je primenjujemo*/
				toggle(i,j);
				/*ako je bio zaledjeni bit, to znaci da je promenom
				izgubljen alel (0 ili 1), pa moramo da ga vratimo nazad, tj.
				ne smemo da ga mutiramo*/
				//if(frozenBit(j))
				//	toggle(i,j);
			}
		}
	}
}

void fitness(chrom *chr){
	(*chr).fit=(*chr).obj;
}

void freeAll(){
	int i;
	for(i=0; i<POPSIZE; i++){
		free(poptemp[i].z);
		free(poptemp[i].lbit);	
		free(popcurrent[i].z);
		free(popcurrent[i].lbit);
	}
	free(prob->l);
	free(prob);
	free(visited);
}

char * strrstr(const char *str, const char *strSearch){
	char *ptr, *last=NULL;
	ptr = (char*)str;
	while((ptr = strstr(ptr, strSearch))) last = ptr++;
	return last;
}

int main(int argc, char* argv[])
{

	unsigned int randSeed;
	int numOfGen;	
	int objhardness;
	float grphdens;
	if(argc>5){
		randSeed=atoi(argv[2]);
		numOfGen=atoi(argv[3]);
		objhardness=atoi(argv[4]);
		grphdens=atof(argv[5]);
		ga(argv[1],randSeed,numOfGen,objhardness,grphdens);
	}else{
		fprintf(stderr, "Upotreba: <prog.exe> <ulFajl> <randSeed> <brGeneracija> <tezinaFjeCilja [0-N]> <gustinaGrafa [0,1]>");
		exit(0);
	}
	return 0;
}


